1000	%title "skeleton_w_fms_xxx.bas"
	%ident					  "100.3"
	declare string constant k_version	= "100.3"			,&
				k_program	= "Skeleton_with_FMS"		! <--- remarks (column 81)
	!=================================================================================================
	! title  : skeleton_w_fms_100.bas (skeleton with FMS)
	! author : Neil Rieck
	! created: 2004.06.14
	! purpose: an introductory demo to show how to use VMS-BASIC with DEC-RMS (ISAM) and DEC-FMS
	! build  : @ SKELETON_W_FMS.COMTXT 100
	! notes  :
	! 0) Typical FMS programs usually load FMS forms from a forms library. Back in the VAX days (when
	!    our systems was under-powered while over-worked, we developed this technique of compiling
	!    FMS forms to object files, then linking them to the executable)
	! 1) this BASIC program worked properly on OpenVMS-VAX and OpenVMS-Alpha
	! 2) it still works properly on OpenVMS-Itanium
	! 3) it fails on OpenVMS-X86 (%BAS-F-MEMMANVIO, Memory management violation) but can be fixed.
	!    Search for FUBARX86a
	! 4) CAVEAT: due to a bug in compiler: VSI BASIC x86-64 X1.8-005-50Y3Q
	!	     you must LINK with this switch: /segment=code=p0
	! 5) CAVEAT: with compiler: VSI BASIC x86-64 V1.9-001
	!            you no longer need to link with /segment=code=p0
	! history:
	! ver who when     what
	! --- --- -------- -------------------------------------------------------------------------------
	! 100 NSR 20040614 1. original code (derived from "skeleton_w/o_fms_xxx.bas")
	!     NSR 20170712 2. a few tweaks after the move to Itanium
	!     NSR 20240430 3. debugging a problem on OpenVMS-X86
	!=================================================================================================
	! Build Method:
	!
	!	$ bas/real=double	skeleton_w_fms_100
	!	$ fms/obj		csmis$flb:skeleton_w_fms_100.frm	/out=skeleton_w_fms_100_frm.obj
	!	$ fms/vec		csmis$flb:skeleton_w_fms_100.frm	/out=skeleton_w_fms_100_Vec.obj
	!	$ link/contig		skeleton_w_fms_100	, 	-
	!				skeleton_w_fms_100_Frm	,	-	
	!				skeleton_w_fms_100_Vec
	!
	! Notes:
	! 1. I had previously created a form in directory "csmis$flb:" called "skeleton_w_fms_100.frm"
	! 2. The commands have been placed into a DCL script named "skeleton_w_fms.com"
	! 3. FMS-CAVEAT-1: some FMS-based programs load forms at run-time from a form library. This
	!    program skips that step by creating an object file which is then included by the linker
	! 4. FMS-CAVEAT-2: not all FMS forms contain vector data
	!=================================================================================================
	option type=explicit							! no kid stuff
	set no prompt								! no automatic '?' on input
	!
	!	constants (usually prefixed with 'k_')
	!
	declare string constant	k_fs1$		= "Skeleton.dat"	,	! the ISAM data file		&
				k_provinces$	= "ON QC NB BC"			!
	!
	! DCL command to view file sizes:
	!	$fms/directory/full Skeleton_w_fms_100.frm
	!
	declare long	constant k_fsize1% = 1000				!
	declare string	constant k_fname1$ = "Skeleton_w_fms_100"		!
	!
	!	home brewed functions and includes
	!
!~	%include "[.inc]device_controls.inc"					x VT + LA device control codes
!~	%include "[.inc]fms_fdvdef.inc"						x get FMS constants
	%include "device_controls.inc"						! VT + LA device control codes
	%include "fms_fdvdef.inc"						! get FMS constants
	!
	external sub trace (string)						!
	external string function wcsm_dt_stamp					!
	!
	!	variable declarations
	!
	declare string	command$,						!&
			choice$,						!&
			my_data$,						!&
			junk$,							!&
			temp$,							!&
		double	temp,							!&
		long	i%,							!&
			j%,							!&
			junk%,							!&
			temp%,							!&
			handler_error%,						!&
			rec_count%						!
	!
	!	This map defines the record layout of our file.
	!	Note that I've used a prefix of "d01_" which means that I will almost always open the assicated file on channel #1
	!
	map(d_skel) string	d01_whole_record	= 132,			!&
				d01_align		= 0			! to enforce align checks by the linker
	!
	!	this map overlays the previous map of the same name (so we want to ensure that it is the same size)
	!
	map(d_skel) string	d01_first_name		= 15	,		! 15	&
				d01_last_name		= 15	,		! 30	&
				d01_address		= 30	,		! 60	&
				d01_city		= 20	,		! 80	&
				d01_province		=  2	,		! 82	&
				d01_telephone		= 10	,		! 92	&
				d01_fill$		= 40	,		!132	need room to grow			&
				d01_align		= 0			!to enforce align checks by the linker
	!
	!	This map defines the record layout of our FMS form
	!	Note that I've used a prefix of "f01_" which means that data used on this form usully is associated 
	!		with a file opened on channel #1
	!
	map(f_skel) string	f01_whole_record	= 92,			!&
				f01_align		= 0			! to enforce align checks by the linker
	!
	!	this map overlays the previous map of the same name (so we want to ensure that it is the same size)
	!
	map(f_skel) string	f01_first_name		= 15	,		! 15	&
				f01_last_name		= 15	,		! 30	&
				f01_address		= 30	,		! 60	&
				f01_city		= 20	,		! 80	&
				f01_province		=  2	,		! 82	&
				f01_telephone		= 10	,		! 92	&
				f01_align		= 0			!to enforce align checks by the linker
	!
	map(wksp1)	string	WorkSpace1	= 12%				! work space 4
	map(tca)	string	TCA		= 12%				! terminal control area
	!
	!	<<< FMS Initialize >>>
	!
	call fdv$aterm( TCA by desc, 12%, 5%,,,16% )				! attach terminal
	call fdv$awksp(	 WorkSpace1 by desc	,k_fsize1%	)		! attach WorkSpace 1
	call fdv$load(	k_fname1$	)					! load form into WorkSpace
	!
	!========================================================================================================================
	!	main
	!========================================================================================================================
	main:
	!
	!	This is a simple text-based menu. I could have used an FMS form here but I wanted to demo the steps necessary
	!	for cleaning up the screen when dealing with displays created by a mixture of FMS and BASIC-PRINT
	!
	call fdv$clear								! 
	print vt$clear + vt$home + k_program +"_"+ k_version			! print title
	print string$(len(k_program +"_"+ k_version), asc("="))			! underline it
	print "  1. display records (with FMS)"
	print "  2. add a record    (w/o  FMS)"
	print "  3. delete a record (w/o  FMS)"
	print "  9. exit"
	print "Choice? (1-3,9) ";
	input command$
	command$ = edit$(command$,32+2)						! upcase, no white space
	command$ = left$(command$,1)						!
	select command$
	    case "1"
		gosub display_records_via_fms
		goto main
	    case "2"
		gosub add_records
		goto main
	    case "3"
		gosub delete_records
		goto main
	    case "9"
		goto fini
	    case else
		print vt$error + "Oops, unsupported command";			! print error text on line #24
		sleep 1
		print vt$error + "";						! clear line #24
		goto main
	end select
	!
	!========================================================================================================================
	!	<<< display records >>>
	!========================================================================================================================
	display_records_via_fms: 
	!
	!	note: fdv$clear will only clear what FMS knows about so if something else has been writing to the screen (like
	!		the text-based menu above), then you need to do a hardware clear to get it all
	!
	call fdv$clear								! clear FMS buffer
	print vt$clear;								! clear VT hardware (just in case)
	call fdv$dispw								! display form work space
%let %FUBARX86a=1								! 0=broken; 1=work around
%if  %FUBARX86a=0								!
%then
	call trace("display - step 004 original")				!
	call fdv$putal("")							! blank all fields on the form (this fails)
%else
	call trace("display - step 004 workaround")				! 
        f01_whole_record = ""							! blank all fields on the map
	call fdv$putal(f01_whole_record)					! blank all fields on the form (this works)
%end %if
	call trace("display - step 005")
	gosub open_data_file							!
	goto display_records_exit	if handler_error% <> 0
	call fdv$getdl(choice$,junk%,23%, "Display by: N/ame, C/ity, P/rovince, T/elephone? (default=N/ame) ")
	choice$ = edit$(choice$,32+2)
	when error in
	    select choice$
		case "C"							! C/ity
		    find #1, key#3 nxeq " "					!
		case "P"							! P/rovince
		    find #1, key#4 nxeq " "					!
		case "T"							! T/elephone
		    find #1, key#2 nxeq " "					!
		case "N"							! N/ame
		    find #1, key#0 nxeq " "					!
		case else							! N/ame (default)
		    call fdv$putl("-i- defaulting to N/ame"+ bel, 24%)
		    sleep 1
		    find #1, key#0 nxeq " "
	    end select
	    while 1
		get #1, regardless						! read w/o lock
		!
		!	xfer and display method #1
		!
		f01_first_name	= d01_first_name				! xfer to the form buffer
		f01_last_name	= d01_last_name					!
		f01_address	= d01_address					!
		f01_city	= d01_city					!
		f01_province	= d01_province					!
		f01_telephone	= d01_telephone					!
		call trace("display - step 006")
		call fdv$putal(f01_whole_record)				! display all fields in one operation
		call trace("display - step 007")
		!
		call fdv$getdl(junk$,junk%,23%,"More? (Y/N,default=Y) ")
		select left$(edit$(junk$,32+2),1)				!
		    case "N"
			cause error 11						! force EOF
		end select
	    next
	use
	    handler_error% = err
	end when
	select handler_error%
	    case 0,								! no error					&
		 11,								! End Of File					&
		 134								! Record Not Found
		call fdv$getdl(junk$,junk%,23%, "hit <enter>...")
	    case else
		call fdv$getdl(junk$,junk%,23%, "-e-Error: "+ str$(handler_error%) + " while reading data. Hit <enter>")
	end select
	!
	display_records_exit:
	return
	!
	!========================================================================================================================
	!	<<< add records >>>
	!========================================================================================================================
	add_records:
	print cr + lf + "Add Records" + cr + lf + 				&
			"==========="
	gosub open_data_file
	goto add_records_exit	if handler_error% <> 0
	when error in
	    d01_whole_record = ""						! start with a clean slate (init the buffer)
	    !
	    input "Last Name  ? ";junk$
	    d01_last_name = edit$(junk$, 128+32+16+8)				! trailing, upcase, compress, leading
	    !
	    input "First Name ? ";junk$
	    d01_first_name = edit$(junk$, 128+32+16+8)				! trailing, upcase, compress, leading
	    !
	    input "Address    ? ";junk$
	    d01_address = edit$(junk$, 128+32+16+8)				! trailing, upcase, compress, leading
	    !
	    input "City       ? ";junk$
	    d01_city = edit$(junk$, 128+32+16+8)				! trailing, upcase, compress, leading
	    !
	    get_prov:
	    input "Province   ? ";junk$
	    junk$ = edit$(junk$, 128+32+16+8)					! trailing, upcase, compress, leading
	    if pos(k_provinces$, junk$, 1) = 0 then
		print "-e- not in the legal list of provinces ("+ k_provinces$ +")"
		goto get_prov
	    end if
	    d01_province = junk$
	    !
	    get_telephone:
	    input "Telephone  ? ";junk$
	    junk$ = edit$(junk$, 128+32+16+8)					! trailing, upcase, compress, leading
	    when error in
		temp = real(junk$,DOUBLE)
		junk% = 0
	    use
		junk% = err
	    end when
	    if junk% <> 0 then
		print "-e- non-numeric character detected in telephone field"
		goto get_telephone
	    end if
	    select len(junk$)
		case 0, 10
		case else
		    print "-e- telephone nust be BLANK or 10 characters"
		    goto get_telephone
	    end select
	    d01_telephone = junk$
	    !
	    !	now give the user a chance to either save or discard his input
	    !
	    print
	    print "Name     : "+	edit$(d01_last_name,128)		!&
					+", "+					!&
					edit$(d01_first_name,128)		!
	    print "Address  : "+	d01_address
	    print "City     : "+	d01_city
	    print "Province : "+	d01_province
	    print "Telephone: "+	d01_telephone
	    input "Save this record? (Y/N, default=N) ";choice$
	    select left$(edit$(choice$,32+2),1)
		case "Y"
		    print "-i-Saving record data"
		    put #1
		case else
		    print "-i-Discarding record data"
	    end select
	    !
	    handler_error% = 0							! signal all is well
	use
	    handler_error% = err
	end when
	select handler_error%
	    case 0								! no error
	    case 11								! End Of File
	    case 134								! Record Not Found
	    case else
		print "-e-Error: "+ str$(handler_error%) + " while writing data"
		print "   Text : "+ ert$(handler_error%)
	end select
	print "hit <enter> to continue...";
	input junk$
	add_records_exit:
	return
	!
	!========================================================================================================================
	!	<<< delete records >>>
	!========================================================================================================================
	delete_records:
	print vt$clear								! clear VT hardware 
	call fdv$clear								! clear FMS buffer
	call fdv$dispw								! display form work space
	call fdv$putal("")
	gosub open_data_file
	goto delete_records_exit	if handler_error% <> 0
	call fdv$getdl(choice$,junk%,23%,"DELETE: Lookup by: N/ame(last), C/ity, P/rovince, T/elephone? (default=exit) ")
	choice$ = edit$(choice$,32+2)
	goto delete_records_exit	if pos("NCPT",choice$,1) = 0		! command not in list so exit
	!
	call fdv$getdl(my_data$,junk%,23%,"DELETE: enter one or more characters? ")
	my_data$ = edit$(my_data$,128+32+16+8)					! trailing, upcase, compress, collapse
	my_data$ = " " if my_data$ = ""						! don't allow a null string here
	when error in
	    select choice$
		case "C"							! C/ity
		    find #1, key#3 nxeq my_data$				!
		case "P"							! P/rovince
		    find #1, key#4 nxeq my_data$				!
		case "T"							! T/elephone
		    find #1, key#2 nxeq my_data$				!
		case "N"							! N/ame
		    find #1, key#0 nxeq my_data$				!
		case else							! default (shouldn't get this far)
		    print "-e- oops, programmer error." + bel			!
		    goto delete_records_exit 
	    end select
	    while 1
		get #1								! read w/lock
		!
		!	note: since I used the phrase "nxeq" in the previous FIND statement,
		!	it is possible that RMS may overshoot the record that I thought was present but was actually missing.
		!	So make sure the record data matches our search data.
		!
		if my_data$ <> " " then						! if not using a blank (but not null) key
		    select choice$
			case "C"						! C/ity
			    cause error 155 if my_data$ <> left$(d01_city,len(my_data$))
			case "P"						! P/rovince
			    cause error 155 if my_data$ <> left$(d01_province,len(my_data$))
			case "T"						! T/elephone
			    cause error 155 if my_data$ <> left$(d01_telephone,len(my_data$))
			case "N"						! N/ame
			    cause error 155 if my_data$ <> left$(d01_last_name,len(my_data$))
			case else						! default (shouldn't get this far)
			    print "-e- oops, programmer error." + bel		!
			    goto delete_records_exit 
		    end select
		end if
		!
		!	at this point, we're sitting on the record with a lock in place and ready to DELETE
		!
		!	xfer and display method #2
		!
		call fdv$put(d01_last_name	,"F$LAST_NAME"		)
		call fdv$put(d01_first_name	,"F$FIRST_NAME"		)
		call fdv$put(d01_address	,"F$ADDRESS"		)
		call fdv$put(d01_city		,"F$CITY"		)
		call fdv$put(d01_province	,"F$PROVINCE"		)
		call fdv$put(d01_telephone	,"F$TELEPHONE"		)
		!
		call fdv$getdl(junk$,junk%,23%, "O.K. to delete this record? (Y/N/S, default=N) ")
		select left$(edit$(junk$,32+2),1)
		    case "Y"							!
			print "-i-Deleting record data"
			delete #1
		    case "S"
			unlock #1						! don't leave a lock on this record
			goto delete_records_exit
		    case else
			call fdv$putl("-i-skipping delete",24%)
			sleep 1
		end select
		call fdv$putl("-i- reading next record (in selected key)",24%)
	    next
	use
	    handler_error% = err
	end when
	select handler_error%
	    case 0								! no error
		call fdv$getdl(junk$,junk%,23%, "hit <enter> to continue...")
	    case else
		call fdv$getdl(junk$,junk%,23%, "-e-Error: "+ str$(handler_error%) + " while reading data. Hit <enter>")
	end select
	!
	delete_records_exit:
	return
	!
	!========================================================================================================================
	!	<<< open data file >>>
	!========================================================================================================================
	open_data_file:
	when error in
	    open k_fs1$ as #1							&
		,organization indexed						&
		,map d_skel							&
		,primary key (d01_last_name, d01_first_name)			&
		,alternate key d01_last_name	duplicates changes		&
		,alternate key d01_telephone	duplicates changes		&
		,alternate key d01_city		duplicates changes		&
		,alternate key d01_province	duplicates changes		!
	    handler_error% = 0							! signal all is well
	use
	    handler_error% = err
	end when
	select handler_error%
	    case 0
	    case else
		call fdv$getdl(junk$,junk%,23%,"-e-Error: "+ str$(handler_error%) + " while opening data file. Hit <enter>")
	end select
	return
	!
	!========================================================================================================================
	!	Adios
	!========================================================================================================================
	fini:
	print "Adios..."	
32000	end									! our program ends here
	!========================================================================================================================
	!	external functions and sub-programs are pulled-in here (after program end)
	!========================================================================================================================
32080	sub trace (string inbound)
	option   type=explicit							! cuz tricks are for kids...
	external string function wcsm_dt_stamp
	when error in
	    open "skeleton_w_fms_trace.txt" as #2, access append, allow write
	    print #2, "-i-"+ wcsm_dt_stamp +" msg: "+ inbound
	    close #2
	use
	    print "-e-error: ";str$(err);" during write to #2"
	end when
	end sub
	!
32100	function string Wcsm_DT_Stamp
	!===================================================================================================================
	! Title  : Wcsm_DT_Stamp_100?.inc
	! Author : Neil S. Rieck
	! Purpose: an external function to return a y2k compliant system time in the form ccyymmddhhmmss (14 chars)
	! Notes  : all our programs call this function so optimizations here will speed up the whole system
	! History:
	! 100  NSR 911229 1. original work
	!===================================================================================================================
	option   type=explicit							! cuz tricks are for kids...
	!
	%include "starlet" %from %library "sys$library:basic$starlet"           !
	!
	declare  long sys_status
	!
	!	this map is required for the call to sys$asctim (format: 19-JUN-1998 23:59:59.1)
	!
	map (WcsmDTStamp0)	string	Sys_buf_22	= 22
	map (WcsmDTStamp0)	string	Sys_day		=  2,	!	&
					Sys_dash1	=  1,	!-	&
					Sys_month	=  3,	!	&
					Sys_dash2	=  1,	!-	&
					Sys_year	=  4,	!	&
					Sys_space	=  1,	!	&
					Sys_Hour	=  2,	!	&
					Sys_colon1	=  1,	!:	&
					Sys_Minute	=  2,	!	&
					Sys_colon2	=  1,	!:	&
					Sys_Second	=  2,	!	&
					Sys_period	=  1,	!.	&
					Sys_Tenth	=  1	!
	!
	!	map for Wcsm date (output)
	!
	map (WcsmDTStamp1)	string	Wcsm_buf_14	= 14	!
	map (WcsmDTStamp1)	string	Wcsm_year	=  4,	!	&
					Wcsm_month	=  2,	!	&
					Wcsm_day	=  2,	!	&
					Wcsm_Hour	=  2,	!	&
					Wcsm_Minute	=  2,	!	&
					Wcsm_Second	=  2
	map (WcsmDTStamp1)	string	Wcsm_year	=  4,	!	&
					Wcsm_month_tens	=  1,	!	&
					Wcsm_month_ones	=  1,	!	&
					Wcsm_day_tens	=  1,	!	&
					Wcsm_day_ones	=  1,	!	&
					Wcsm_Hour	=  2,	!	&
					Wcsm_Minute	=  2,	!	&
					Wcsm_Second	=  2
	!
	!	string constants
	!					  00000000011111111112222222222333333333
	!					  12345678901234567890123456789012345678
	declare string constant k_month_names$ = "XXJANFEBMARAPRMAYJUNJULAUGSEPOCTNOVDEC"
	!					  ||
	!					  ++-- so I don't have to provide an offset in pos()
	declare string constant my_space = '32'C
	!
	!	<<< function 'code' starts here >>>
	!
	when error in
		!
		sys_status = sys$asctim(,Sys_buf_22,,)				! get ASCII time into sys_buf_22
!~~~		if (sys_status and 7%) <> 1% then cause error 11		x  not required - call will never fail
		!
		!	transfer data from one map to the other
		!	
		Wcsm_year	= Sys_year					!
!~~~	rset	Wcsm_month	= str$( pos(k_month_names$,Sys_Month,1%) / 3%)	x					bf_100f
		Wcsm_day	= Sys_day					!
		Wcsm_hour	= Sys_hour					!
		Wcsm_minute	= Sys_minute					!
		Wcsm_second	= Sys_second					!
		!
		declare long temp%						!					bf_100f
		temp% = pos(k_month_names$,Sys_Month,1%) / 3%			! compute month number			bf_100f
		if temp% < 10% then						! if less than 10...			bf_100f
		    Wcsm_month_ones	= str$(temp%)				! ...then this goes into ONES		bf_100f
		    Wcsm_month_tens	= "0"					! ...and this goes into TENS		bf_100f
		else								! else >= 10				bf_100f
		    Wcsm_month		= str$(temp%)				!					bf_100f
		end if
		!
		!	make sure there are no spaces in the TENS area of our mapped variables (pad with '0' if necessary)
		!
!~~~		Wcsm_month_tens = "0"	if Wcsm_month_tens	= my_space	x disabled - see above code		bf_100f
		Wcsm_day_tens	= "0"	if Wcsm_day_tens	= my_space	!
		!
		!	now pass result back to caller
		!
		Wcsm_DT_Stamp = Wcsm_Buf_14					! this is it folks
	use
		Wcsm_DT_Stamp = ""						! error so return blank
	end when
	!
	END Function
	!======================================================================= the last line
