1000	%title "skeleton_w_fms_xxx.bas"
	%ident					  "101.2"
	declare string constant k_version	= "101.2"			,&
				k_program	= "Skeleton_with_FMS"		! <--- remarks (column 81)
	!=================================================================================================
	! title  : skeleton_w_fms_100.bas (skeleton with FMS)
	! author : Neil Rieck
	! created: 2004.06.14
	! purpose: an introductory demo to show how to use VMS-BASIC with DEC-RMS (ISAM) and DEC-FMS
	! build  : @ SKELETON_W_FMS.COMTXT 101
	! notes  :
	! 0) Typical FMS programs usually load FMS forms from a forms library. Back in the VAX days (when
        !    our system was under-powered while over-worked, we developed this technique of compiling
        !    FMS forms to object files, then linking them to the executable)
	! 1) this BASIC program works properly on OpenVMS-VAX and OpenVMS-Alpha and OpenVMS-Itanium
	! 2) it fails on OpenVMS-X86 (%BAS-F-MEMMANVIO, Memory management violation) but can be fixed.
	!    Search for FUBARX86a
	! 3) CAVEAT: due to a bug in compiler: VSI BASIC x86-64 X1.8-005-50Y3Q
	!            you must LINK with this switch: /segment=code=p0
	! 4) CAVEAT: with compiler: VSI BASIC x86-64 V1.9-001
	!            you no longer need to link with /segment=code=p0
	! 5) If it looks like I pasted in a bunch of routines from another program, then your
	!    observation would be correct. Some of this code is 40 years old (VAX)
	! 6) With VAX + ALpha you may load a form multiple times into a workspace
	!    With X86 you may only load a form once into a workspace (throws error:
	!	%BAS-F-MEMMANVIO, Memory management violation) There are two workarounds
	!	1) always preceed fdv$load with fdv$awksp (Search for FUBARX86b)
	!	-or-
	!	2) Introduce new logic to ensure we only load once.
	! history:
	! ver who when     what
	! --- --- -------- -------------------------------------------------------------------------------
	! 100 NSR 20040614 1. original code (derived from "skeleton_w/o_fms_xxx.bas")
	!     NSR 20170712 2. a few tweaks
	!     NSR 20240430 3. debugging a problem on OpenVMS-X86 (see FUBARX86a)
	!     NSR 20240830 4. retesting with VSI BASIC x86-64 V1.9-001
	! 101 NSR 20240830 1. added command 4 (add a record (with FMS)
	!                  2. now load form skeleton_w_fms_101 (linked to help form: skeleton_fms_help)
	!=================================================================================================
	! Build Method (Itanium + Alpha):
	!
	!	$ bas/real=double	skeleton_w_fms_101
	!	$ fms/obj		csmis$flb:skeleton_w_fms_101.frm	/out=skeleton_w_fms_101_frm.obj
	!	$ fms/vec		csmis$flb:skeleton_w_fms_101.frm	/out=skeleton_w_fms_101_Vec.obj
	!	$ link/contig		skeleton_w_fms_101	, 	-
	!				skeleton_w_fms_101_Frm	,	-	
	!				skeleton_w_fms_101_Vec
	! Notes:
	! 1. I had previously created a form in directory "csmis$flb:" called "skeleton_w_fms_100.frm"
	! 2. The commands have been placed into a DCL script named "skeleton_w_fms.com"
	! 3. FMS-CAVEAT-1: some FMS-based programs load forms at run-time from a form library. This
	!    program skips that step by creating an object file which is then included by the linker
	! 4. FMS-CAVEAT-2: not all FMS forms contain vector data
	!=================================================================================================
	option type=explicit							! no kid stuff
	set no prompt								! no automatic '?' on input
	!
	!	constants (usually prefixed with 'k_')
	!
	declare string constant	k_fs1$		= "Skeleton.dat"	,	! the ISAM data file	&
				k_provinces$	= "AB BC MB NB NL NT NS NU ON PE QC SK YT"
	!
	! DCL command to view file sizes:
	!	$fms/directory/full Skeleton_w_fms_101.frm
	!
	declare long	constant k_fsize1% = 2000				!
	declare string	constant k_fname1$ = "Skeleton_w_fms_101"		! this name is found in the included form
	!
	!	home brewed functions and includes
	!
!~	%include "[.inc]device_controls.inc"					x VT + LA device control codes
!~	%include "[.inc]fms_fdvdef.inc"						x get FMS constants
	%include "device_controls.inc"						! VT + LA device control codes
	%include "fms_fdvdef.inc"						! get FMS constants
	!
	external sub trace (string)						!
	external string function wcsm_dt_stamp					!
	external long   function funct_fms_nav( long, long, long, long, long, dim(), dim() )
	external long   function find_field_number(string dim() by ref, long by ref, string)	!
	!
	!	variable declarations
	!
	declare string	command$,						!&
			choice$,						!&
			my_data$,						!&
			junk$,							!&
			temp$,							!&
		double	temp,							!&
		long	i%,							!&
			j%,							!&
			junk%,							!&
			temp%,							!&
			handler_error%,						!&
			rec_count%						!
	!
	!	This map defines the record layout of our file.
	!	Note that I've used a prefix of "d01_" which means that I will almost always open the assicated file on channel #1
	!
	map(d_skel) string	d01_whole_record	= 132,			!&
				d01_align		= 0			! to enforce align checks by the linker
	!
	!	this map overlays the previous map of the same name (so we want to ensure that it is the same size)
	!
	map(d_skel) string	d01_first_name		= 15	,		! 15	&
				d01_last_name		= 15	,		! 30	&
				d01_address		= 30	,		! 60	&
				d01_city		= 20	,		! 80	&
				d01_province		=  2	,		! 82	&
				d01_telephone		= 10	,		! 92	&
				d01_fill$		= 40	,		!132	need room to grow			&
				d01_align		= 0			!to enforce align checks by the linker
	!
	!	This map defines the record layout of our FMS form
	!	Note that I've used a prefix of "f01_" which means that data used on this form usully is associated 
	!		with a file opened on channel #1
	!
	map(f_skel) string	f01_whole_record	= 92,			!&
				f01_align		= 0			! to enforce align checks by the linker
	!
	!	this map overlays the previous map of the same name (so we want to ensure that it is the same size)
	!
	map(f_skel) string	f01_first_name		= 15	,		! 15	&
				f01_last_name		= 15	,		! 30	&
				f01_address		= 30	,		! 60	&
				f01_city		= 20	,		! 80	&
				f01_province		=  2	,		! 82	&
				f01_telephone		= 10	,		! 92	&
				f01_align		= 0			!to enforce align checks by the linker
	!
	map(wksp1)	string	WorkSpace1	= 12%				! work space 1 (for form 1)
	map(wksp1)	string	WorkSpace2	= 12%				! work space 2 (for form 2)
	map(wksp1)	string	WorkSpace3	= 12%				! work space 3 (for form 3)
	map(tca)	string	TCA		= 12%				! terminal control area
	!
	declare		long	form_init_flags%		,		!			&
				desired_form%			,		!			&
				curr_wksp%			,		!			&
				g_max_fields1			,		!			&
				fms_fld_enable0			,		!			&
			string	field_name$			,		!			&
				field_data$					!
	! params for form #1
	declare		long	fld1				,		! current field number	&
				direction1			,		! current direction	&
				term1				,		! current terminator	&
				field_limit_lo1			,		! field limit: low	&
				field_limit_hi1			,		! field limit: high	&
				fms_fld_up1(20)			,		! up   pointer		&
				fms_fld_dwn1(20)		,		! down pointer		&
			string	fms_fld_name1(20)				! field name
	!
	!	<<< FMS Initialize >>>
	!
	call fdv$aterm( TCA by desc, 12%, 5%,,,16% )				! attach terminal
	call fdv$awksp(	 WorkSpace1 by desc	,k_fsize1%	)		! attach WorkSpace 1
	call fdv$load( k_fname1$)						! load form into work space
	call fdv$ndisp								! detach from WorkSpace
	curr_wksp% = 0								! good housekeeping
	!
	!========================================================================================================================
	!	main
	!========================================================================================================================
	main:
	!
	!	This is a simple text-based menu. I could have used an FMS form here but, I wanted to demo the steps
	!	necessary for cleaning up the screen when dealing with displays created by a mixture of FMS and BASIC-PRINT
	!-----------------------------------------------------------------------
	!	We can only write to the screen with BASIC when the workspace is detached
	!
	if curr_wksp% <> 0 then							!
	    call fdv$clear							!
	    call fdv$ndisp							!
	    curr_wksp% = 0							!
	end if									!
	print vt$clear + vt$home + k_program +"_"+ k_version			! print title
	print string$(len(k_program +"_"+ k_version), asc("="))			! underline it
	print "  1. display records (with FMS)"
	print "  2. add a record    (w/o  FMS)"
	print "  3. delete a record (w/o  FMS)"
	print "  4. add a record    (with FMS simple)"
	print "  5. add a record    (with FMS fancy)"
	print "  9. exit"
	print "Choice? (1-3,9) ";
	input command$
	command$ = edit$(command$,32+2)						! upcase, no white space
	command$ = left$(command$,1)						!
	select command$
	    case "1"
		gosub display_records_via_fms
		goto main
	    case "2"
		gosub add_records
		goto main
	    case "3"
		gosub delete_records
		goto main
	    case "4"
		gosub add_records_w_fms_simple
		goto main
	    case "5"
		gosub add_records_w_fms_fancy
		goto main
	    case "9"
		goto fini
	    case else
		print vt$error + "Oops, unsupported command";			! print error text on line #24
		sleep 1
		print vt$error + "";						! clear line #24
		goto main
	end select
	!
	!========================================================================================================================
	!	<<< display records >>>
	!========================================================================================================================
	display_records_via_fms:
	!
	!	note: fdv$clear will only clear what FMS knows about so if something else has been writing to the screen (like
	!		the text-based menu above), then you need to do a hardware clear to get it all
	!
	print vt$clear + vt$home;						! clear VT hardware (just in case)
	call fdv$swksp( WorkSpace1 by desc )
	curr_wksp% = 1
	call fdv$clear
%let %FUBARX86b = 1
%if  %FUBARX86b = 1 %then
	call fdv$awksp(	 WorkSpace1 by desc	,k_fsize1%	)		! re-attach WorkSpace 1
%end %if
	call fdv$load(	k_fname1$	)					! load form into WorkSpace
	call fdv$dispw								! display form work space
        f01_whole_record = ""							! blank all fields on the map for this form
%let %FUBARX86a=1								! 0=broken; 1=work around
%if  %FUBARX86a=0								!
%then										!
	call trace("display - step 004 original code for Alpha + Itanium")	!
	call fdv$putal("")							! blank all fields on the form (this fails)
%else										!
	call trace("display - step 004 workaround for x86")			! 
        f01_whole_record = ""							! blank all fields on the map for this form
	call fdv$putal(f01_whole_record)					! blank all fields on the form (this works)
%end %if
	call trace("display - step 005")
	gosub open_data_file							!
	goto display_records_exit	if handler_error% <> 0
	call fdv$getdl(choice$,junk%,23%, "Display by: N/ame, C/ity, P/rovince, T/elephone? (default=N/ame) ")
	choice$ = edit$(choice$,32+2)
	when error in
	    select choice$
		case "C"							! C/ity
		    find #1, key#3 nxeq " "					!
		case "P"							! P/rovince
		    find #1, key#4 nxeq " "					!
		case "T"							! T/elephone
		    find #1, key#2 nxeq " "					!
		case "N"							! N/ame
		    find #1, key#0 nxeq " "					!
		case else							! N/ame (default)
		    call fdv$putl("-i- defaulting to N/ame"+ bel, 24%)
		    sleep 1
		    find #1, key#0 nxeq " "
	    end select
	    while 1
		get #1, regardless						! read w/o lock
		!
		!	xfer and display method #1
		!
		f01_first_name	= d01_first_name				! xfer to the form buffer
		f01_last_name	= d01_last_name					!
		f01_address	= d01_address					!
		f01_city	= d01_city					!
		f01_province	= d01_province					!
		f01_telephone	= d01_telephone					!
		call trace("display - step 006")
		call fdv$putal(f01_whole_record)				! display all fields in one operation
		call trace("display - step 007")
		!
		call fdv$getdl(junk$,junk%,23%,"More? (Y/N,default=Y) ")
		select left$(edit$(junk$,32+2),1)				!
		    case "N"
			cause error 11						! force EOF
		end select
	    next
	use
	    handler_error% = err
	end when
	select handler_error%
	    case 0,								! no error					&
		 11,								! End Of File					&
		 134								! Record Not Found
		call fdv$getdl(junk$,junk%,23%, "hit <enter>...")
	    case else
		call fdv$getdl(junk$,junk%,23%, "-e-Error: "+ str$(handler_error%) + " while reading data. Hit <enter>")
	end select
	!
	display_records_exit:
	return
	!
	!========================================================================================================================
	!	<<< add records >>>
	!========================================================================================================================
	add_records:
	print cr + lf + "Add Records" + cr + lf + 				&
			"==========="
	gosub open_data_file
	goto add_records_exit	if handler_error% <> 0
	when error in
	    d01_whole_record = ""						! start with a clean slate (init the buffer)
	    !
	    input "Last Name  ? ";junk$
	    d01_last_name = edit$(junk$, 128+32+16+8)				! trailing, upcase, compress, leading
	    !
	    input "First Name ? ";junk$
	    d01_first_name = edit$(junk$, 128+32+16+8)				! trailing, upcase, compress, leading
	    !
	    input "Address    ? ";junk$
	    d01_address = edit$(junk$, 128+32+16+8)				! trailing, upcase, compress, leading
	    !
	    input "City       ? ";junk$
	    d01_city = edit$(junk$, 128+32+16+8)				! trailing, upcase, compress, leading
	    !
	    get_prov:
	    input "Province   ? ";junk$
	    junk$ = edit$(junk$, 128+32+16+8)					! trailing, upcase, compress, leading
	    d01_province = junk$						!
	    f01_province = junk$						!
	    if (len(junk$) <> 2) or (pos(k_provinces$,junk$,1) = 0) then
		print "-e-not a legal province ("+ k_provinces$ +")"
		goto get_prov
	    end if
	    !
	    get_telephone:
	    input "Telephone  ? ";junk$
	    junk$ = edit$(junk$, 128+32+16+8)					! trailing, upcase, compress, leading
	    when error in
		temp = real(junk$,DOUBLE)
		junk% = 0
	    use
		junk% = err
	    end when
	    if junk% <> 0 then
		print "-e- non-numeric character detected in telephone field"
		goto get_telephone
	    end if
	    select len(junk$)
		case 0, 10
		case else
		    print "-e- telephone nust be BLANK or 10 characters"
		    goto get_telephone
	    end select
	    d01_telephone = junk$
	    !
	    !	now give the user a chance to either save or discard his input
	    !
	    print
	    print "Name     : "+	edit$(d01_last_name,128)		!&
					+", "+					!&
					edit$(d01_first_name,128)		!
	    print "Address  : "+	d01_address
	    print "City     : "+	d01_city
	    print "Province : "+	d01_province
	    print "Telephone: "+	d01_telephone
	    input "Save this record? (Y/N, default=N) ";choice$
	    select left$(edit$(choice$,32+2),1)
		case "Y"
		    print "-i-Saving record data"
		    put #1
		case else
		    print "-i-Discarding record data"
	    end select
	    !
	    handler_error% = 0							! signal all is well
	use
	    handler_error% = err
	end when
	select handler_error%
	    case 0								! no error
	    case 11								! End Of File
	    case 134								! Record Not Found
	    case else
		print "-e-Error: "+ str$(handler_error%) + " while writing data"
		print "   Text : "+ ert$(handler_error%)
	end select
	print "hit <enter> to continue...";
	input junk$
	add_records_exit:
	return
	!
	!========================================================================================================================
	!	<<< add records (with fms) >>>
	!========================================================================================================================
	add_records_w_fms_simple:
	print vt$clear + vt$home;						! clear VT hardware (just in case)
	call fdv$swksp( WorkSpace1 by desc )
	curr_wksp% = 1
	call fdv$clear
%if  %FUBARX86b = 1 %then
	call fdv$awksp(	 WorkSpace1 by desc	,k_fsize1%	)		! re-attach WorkSpace 1
%end %if
	call fdv$load(	k_fname1$	)					! load form into WorkSpace
	call fdv$dispw								! display form work space
        f01_whole_record = ""							! blank all fields on the map for this form
	call fdv$putal(f01_whole_record)					! update all fields on the form (this works)
	!
	loop1:
	call fdv$putl("Keying-help> TAB:(NextField) Ctrl-H:(PreviousField) ENTER:(Exit Form)" ,23%)		!
	call fdv$getal(f01_whole_record)					!
	!
	gosub open_data_file
	goto add_records_exit_w_fms	if handler_error% <> 0
	call fdv$getdl(junk$,junk%,23%,"okay to validate? (y/n) ")
	goto add_records_exit_w_fms	if edit$(junk$,32+2) <> "Y"
	when error in
	    d01_whole_record = ""						! start with a clean slate (init the buffer)
	    !
	    d01_last_name = edit$(f01_last_name, 128+32+16+8)			! trailing, upcase, compress, leading
	    !
	    d01_first_name = edit$(f01_first_name, 128+32+16+8)			! trailing, upcase, compress, leading
	    !
	    d01_address = edit$(f01_address, 128+32+16+8)			! trailing, upcase, compress, leading
	    !
	    d01_city = edit$(f01_city, 128+32+16+8)				! trailing, upcase, compress, leading
	    !
	    get_prov2:
	    junk$ = edit$(f01_province, 128+32+16+8)				! trailing, upcase, compress, leading
	    f01_province = junk$
	    d01_province = junk$
	    if (len(junk$) <> 2) or (pos(k_provinces$,junk$,1) = 0) then	!
		call fdv$putl("-e-not a legal provinces ("+ k_provinces$ +")", 24%)
		sleep 1
		goto loop1							!
	    end if								!
	    !
	    get_telephone2:
	    junk$ = edit$(f01_telephone, 128+32+16+8)				! trailing, upcase, compress, leading
	    when error in
		temp = real(junk$,DOUBLE)
		junk% = 0
	    use
		junk% = err
	    end when
	    if junk% <> 0 then
		call fdv$putl("-e- non-numeric character detected in telephone field", 24%)
		sleep 1
		goto loop1
	    end if
	    select len(junk$)
		case 0, 10
		case else
		    call fdv$putl("-e- telephone nust be BLANK or 10 characters", 24%)
		    goto loop1
	    end select
	    d01_telephone = junk$
	    !
	    call fdv$getdl(junk$,junk%,23%,"okay to save? (y/n) ")
	    select edit$(junk$,32+2)
   		case "Y"
		    call fdv$putl("-i-Saving record data",24%)
		    put #1
		case else
		    call fdv$putl("-i-Discarding record data",24%)
	    end select
	    sleep 1
	    !
	    handler_error% = 0							! signal all is well
	use
	    handler_error% = err
	end when
	select handler_error%
	    case 0								! no error
	    case 11								! End Of File
	    case 134								! Record Not Found
	    case else
		call fdv$putl("-e-Error: "+ str$(handler_error%) + " while writing data",23%)
		call fdv$putl("-i-Text : "+ ert$(handler_error%),24%)
		sleep 1
	end select
	call fdv$putl("",24%)
	call fdv$getdl(junk$,junk%,23%,"hit <enter> to continue...")
	add_records_exit_w_fms:
	return
	!
	!========================================================================================================================
	!	<<< add records (with fms fancy, er messy) >>>
	!========================================================================================================================
	! Note: This routines fully supports stuff like:
	!		arrow-up (previous field)
	!		arrow-down (next field)
	!		tab (next field)
	!		HELP
	!-----------------------------------------------------------------------
	add_records_w_fms_fancy:
	desired_form% = 1							! we want form 1
	select desired_form%
	    case 1
		select curr_wksp%
		    case 0
			print vt$clear + vt$home;				! clear VT hardware (just in case)
		    case <> 1
			call fdv$clear
			call fdv$ndisp
			curr_wksp% = 0
		end select
		if ( (form_init_flags% and 1%)=0% ) then			! if not yet initialized
		    call fdv$awksp(	WorkSpace1 by desc, k_fsize1% )		! attach WorkSpace 1
		    call fdv$load(	k_fname1$	)			! load form into WorkSpace
		    call fdv$dispw						! display workspace
		    call fdv$retdn( "NUMBER_OF_FIELDS", junk$ )			!
		    g_max_fields1 = integer( junk$ )				!
		    field_limit_lo1 = 1						!
		    field_limit_hi1 = g_max_fields1				!
		    for i% = 1 to g_max_fields1					!
			call fdv$retfo ( i%, Field_name$, junk% )		!
			fms_fld_name1  ( i% ) = Field_name$			!
			call fdv$retdn ("NAV"+str$( i%), junk$)			! eg NAV02 >>> 01,03
			fms_fld_up1    ( i% ) = integer(mid$( junk$,1,2))	! 01
			fms_fld_dwn1   ( i% ) = integer(mid$( junk$,4,2))	! 03
		    next i%							!
		    !
		    form_init_flags% = form_init_flags% or 1%			! show form #1 init'd
		else								!
		    if curr_wksp% <> 1 then					!
			call fdv$swksp( WorkSpace1 by desc)			! switch to workspace
			call fdv$dispw						! display workspace
		    end if							!
		end if								!
		!
		fms_fld_enable0 = -1						! enable all fields
		direction1 = 1							! forward
		curr_wksp% = 1							! show current full screen form
		f01_whole_record = ""						!
		call fdv$putal(f01_whole_record)				! update all fields on the form (this works)
	end select								!
	fld1 = field_limit_lo1							!
	!
	loop1f:
	call fdv$iltrm(1%)							! enable illegal terminators ---vvv
	select fld1								! do range checks
	    case < field_limit_lo1						!
		fld1 = field_limit_hi1						! wrap around
	    case > field_limit_hi1						!
		fld1 = field_limit_lo1						! wrap around
	end select								!
        if (fms_fld_enable0 and (2%^fld1))=0% then                              ! skip disabled fields
            fld1 = fld1 + direction1                                            !
            goto loop1f                                                         ! re-do range checks
        end if                                                                  !
	field_name$ = fms_fld_name1( fld1 )					!
	select field_name$							!
	    case "F$WHATEVER"							! per field prompt
		call fdv$putl("per field hint", ,24%)				!
	    case else								!
		call fdv$putl(field_name$ +" (Up/Down/TAB/BS/Enter/PF2/PF4)" ,24%)
	end select								!
	call fdv$get( Field_data$, term1, Field_name$  )			!
	call fdv$iltrm(0%)                                                      ! disable illegal terminators ---^^^
        goto loop1f	if 1% = funct_fms_nav(  term1                  ,&
                                                direction1             ,&
                                                fld1                   ,&
                                                field_limit_hi1        ,&
                                                field_limit_lo1        ,&
                                                fms_fld_up1()          ,&
                                                fms_fld_dwn1()          )
	call fdv$retal(f01_whole_record)					!
	call fdv$getdl(junk$,junk%,23%,"-?-okay to validate? (y/N) ")		!
	goto add_records_exit_w_fms_fancy	if edit$(junk$,32+2) <> "Y"	!
	!
	gosub open_data_file
	goto add_records_exit_w_fms_fancy	if handler_error% <> 0
	when error in
	    d01_whole_record = ""						! start with a clean slate (init the buffer)
	    !
	    d01_last_name = edit$(f01_last_name, 128+32+16+8)			! trailing, upcase, compress, leading
	    !
	    d01_first_name = edit$(f01_first_name, 128+32+16+8)			! trailing, upcase, compress, leading
	    !
	    d01_address = edit$(f01_address, 128+32+16+8)			! trailing, upcase, compress, leading
	    !
	    d01_city = edit$(f01_city, 128+32+16+8)				! trailing, upcase, compress, leading
	    !
	    get_prov2f:
	    junk$ = edit$(f01_province, 128+32+16+8)				! trailing, upcase, compress, leading
	    f01_province = junk$
	    d01_province = junk$
	    if (len(junk$)<>2) or (pos(k_provinces$,junk$,1) = 0) then		!
		call fdv$putl("-e-not a legal province ("+ k_provinces$ +")", 24%)
		sleep 1
		fld1 = find_field_number(fms_fld_name1(), g_max_fields1, "F$PROVINCE")
		goto loop1f							!
	    end if								!
	    !
	    get_telephone2f:
	    junk$ = edit$(f01_telephone, 128+32+16+8)				! trailing, upcase, compress, leading
	    when error in
		temp = real(junk$,DOUBLE)
		junk% = 0
	    use
		junk% = err
	    end when
	    if junk% <> 0 then
		call fdv$putl("-e- non-numeric character detected in telephone field", 24%)
		sleep 1
		fld1 = find_field_number(fms_fld_name1(), g_max_fields1, "F$TELEPHONE")
		goto loop1f
	    end if
	    select len(junk$)
		case 0, 10
		case else
		    call fdv$putl("-e- telephone nust be BLANK or 10 characters", 24%)
		    sleep 1
		    fld1 = find_field_number(fms_fld_name1(), g_max_fields1, "F$TELEPHONE")
		    goto loop1f
	    end select
	    d01_telephone = junk$
	    !
	    !	now give the user a chance to either save or discard his input
	    !
	    call fdv$getdl(choice$, junk%, 23%, "Okay to save? (Y/N, default=N) ")
	    select left$(edit$(choice$,32+2),1)
		case "Y"
		    call fdv$putl("-i-Saving record data", 23%)
		    put #1
		case else
		    call fdv$putl("-i-Discarding record data", 23%)
	    end select
	    sleep 1
	    !
	    handler_error% = 0							! signal all is well
	use
	    handler_error% = err
	end when
	select handler_error%
	    case 0								! no error
	    case 11								! End Of File
	    case 134								! Record Not Found
	    case else
		call fdv$putl("-e-Error: "+ str$(handler_error%) + " while writing data",23%)
		call fdv$putl("   Text : "+ ert$(handler_error%),24%)
		sleep 1
	end select
	call fdv$putl("", 24%)
	call fdv$getdl(junk$,junk%,23%,"hit <enter> to continue...")
	add_records_exit_w_fms_fancy:
	return
	!
	!========================================================================================================================
	!	<<< delete records >>>
	!========================================================================================================================
	delete_records:
	print vt$clear + vt$home;						! clear VT hardware 
	call fdv$clear								! clear FMS buffer
	call fdv$dispw								! display form work space
	call fdv$putal("")
	gosub open_data_file
	goto delete_records_exit	if handler_error% <> 0
	call fdv$getdl(choice$,junk%,23%,"DELETE: Lookup by: N/ame(last), C/ity, P/rovince, T/elephone? (default=exit) ")
	choice$ = edit$(choice$,32+2)
	goto delete_records_exit	if pos("NCPT",choice$,1) = 0		! command not in list so exit
	!
	call fdv$getdl(my_data$,junk%,23%,"DELETE: enter one or more characters? ")
	my_data$ = edit$(my_data$,128+32+16+8)					! trailing, upcase, compress, collapse
	my_data$ = " " if my_data$ = ""						! don't allow a null string here
	when error in
	    select choice$
		case "C"							! C/ity
		    find #1, key#3 nxeq my_data$				!
		case "P"							! P/rovince
		    find #1, key#4 nxeq my_data$				!
		case "T"							! T/elephone
		    find #1, key#2 nxeq my_data$				!
		case "N"							! N/ame
		    find #1, key#0 nxeq my_data$				!
		case else							! default (shouldn't get this far)
		    print "-e- oops, programmer error." + bel			!
		    goto delete_records_exit 
	    end select
	    while 1
		get #1								! read w/lock
		!
		!	note: since I used the phrase "nxeq" in the previous FIND statement,
		!	it is possible that RMS may overshoot the record that I thought was present but was actually missing.
		!	So make sure the record data matches our search data.
		!
		if my_data$ <> " " then						! if not using a blank (but not null) key
		    select choice$
			case "C"						! C/ity
			    cause error 155 if my_data$ <> left$(d01_city,len(my_data$))
			case "P"						! P/rovince
			    cause error 155 if my_data$ <> left$(d01_province,len(my_data$))
			case "T"						! T/elephone
			    cause error 155 if my_data$ <> left$(d01_telephone,len(my_data$))
			case "N"						! N/ame
			    cause error 155 if my_data$ <> left$(d01_last_name,len(my_data$))
			case else						! default (shouldn't get this far)
			    print "-e- oops, programmer error." + bel		!
			    goto delete_records_exit 
		    end select
		end if
		!
		!	at this point, we're sitting on the record with a lock in place and ready to DELETE
		!
		!	xfer and display method #2
		!
		call fdv$put(d01_last_name	,"F$LAST_NAME"		)
		call fdv$put(d01_first_name	,"F$FIRST_NAME"		)
		call fdv$put(d01_address	,"F$ADDRESS"		)
		call fdv$put(d01_city		,"F$CITY"		)
		call fdv$put(d01_province	,"F$PROVINCE"		)
		call fdv$put(d01_telephone	,"F$TELEPHONE"		)
		!
		call fdv$getdl(junk$,junk%,23%, "O.K. to delete this record? (Y/N/S, default=N) ")
		select left$(edit$(junk$,32+2),1)
		    case "Y"							!
			print "-i-Deleting record data"
			delete #1
		    case "S"
			unlock #1						! don't leave a lock on this record
			goto delete_records_exit
		    case else
			call fdv$putl("-i-skipping delete",24%)
			sleep 1
		end select
		call fdv$putl("-i- reading next record (in selected key)",24%)
	    next
	use
	    handler_error% = err
	end when
	select handler_error%
	    case 0								! no error
		call fdv$getdl(junk$,junk%,23%, "hit <enter> to continue...")
	    case else
		call fdv$getdl(junk$,junk%,23%, "-e-Error: "+ str$(handler_error%) + " while reading data. Hit <enter>")
	end select
	!
	delete_records_exit:
	return
	!
	!========================================================================================================================
	!	<<< open data file >>>
	!========================================================================================================================
	open_data_file:
	when error in
	    open k_fs1$ as #1							&
		,organization indexed						&
		,map d_skel							&
		,primary key (d01_last_name, d01_first_name)			&
		,alternate key d01_last_name	duplicates changes		&
		,alternate key d01_telephone	duplicates changes		&
		,alternate key d01_city		duplicates changes		&
		,alternate key d01_province	duplicates changes		!
	    handler_error% = 0							! signal all is well
	use
	    handler_error% = err
	end when
	select handler_error%
	    case 0
	    case else
		call fdv$getdl(junk$,junk%,23%,"-e-Error: "+ str$(handler_error%) + " while opening data file. Hit <enter>")
	end select
	return
	!
	!========================================================================================================================
	!	Adios
	!========================================================================================================================
	fini:
	print "Adios..."	
32000	end									! our program ends here
	!========================================================================================================================
	!	external functions and sub-programs are pulled-in here (after program end)
	!========================================================================================================================
32010	sub trace (string inbound)
	option   type=explicit							! cuz tricks are for kids...
	external string function wcsm_dt_stamp
	when error in
	    open "skeleton_w_fms_trace.txt" as #2, access append, allow write
	    print #2, "-i-"+ wcsm_dt_stamp +" msg: "+ inbound
	    close #2
	use
	    print "-e-error: ";str$(err);" during write to #2"
	end when
	end sub
	!
32020	function string Wcsm_DT_Stamp
	!===================================================================================================================
	! Title  : Wcsm_DT_Stamp_100?.inc
	! Author : Neil S. Rieck
	! Purpose: an external function to return a y2k compliant system time in the form ccyymmddhhmmss (14 chars)
	! Notes  : all our programs call this function so optimizations here will speed up the whole system
	! History:
	! 100  NSR 911229 1. original work
	!===================================================================================================================
	option   type=explicit							! cuz tricks are for kids...
	!
	%include "starlet" %from %library "sys$library:basic$starlet"           !
	!
	declare  long sys_status
	!
	!	this map is required for the call to sys$asctim (format: 19-JUN-1998 23:59:59.1)
	!
	map (WcsmDTStamp0)	string	Sys_buf_22	= 22
	map (WcsmDTStamp0)	string	Sys_day		=  2,	!	&
					Sys_dash1	=  1,	!-	&
					Sys_month	=  3,	!	&
					Sys_dash2	=  1,	!-	&
					Sys_year	=  4,	!	&
					Sys_space	=  1,	!	&
					Sys_Hour	=  2,	!	&
					Sys_colon1	=  1,	!:	&
					Sys_Minute	=  2,	!	&
					Sys_colon2	=  1,	!:	&
					Sys_Second	=  2,	!	&
					Sys_period	=  1,	!.	&
					Sys_Tenth	=  1	!
	!
	!	map for Wcsm date (output)
	!
	map (WcsmDTStamp1)	string	Wcsm_buf_14	= 14	!
	map (WcsmDTStamp1)	string	Wcsm_year	=  4,	!	&
					Wcsm_month	=  2,	!	&
					Wcsm_day	=  2,	!	&
					Wcsm_Hour	=  2,	!	&
					Wcsm_Minute	=  2,	!	&
					Wcsm_Second	=  2
	map (WcsmDTStamp1)	string	Wcsm_year	=  4,	!	&
					Wcsm_month_tens	=  1,	!	&
					Wcsm_month_ones	=  1,	!	&
					Wcsm_day_tens	=  1,	!	&
					Wcsm_day_ones	=  1,	!	&
					Wcsm_Hour	=  2,	!	&
					Wcsm_Minute	=  2,	!	&
					Wcsm_Second	=  2
	!
	!	string constants
	!					  00000000011111111112222222222333333333
	!					  12345678901234567890123456789012345678
	declare string constant k_month_names$ = "XXJANFEBMARAPRMAYJUNJULAUGSEPOCTNOVDEC"
	!					  ||
	!					  ++-- so I don't have to provide an offset in pos()
	declare string constant my_space = '32'C
	!
	!	<<< function 'code' starts here >>>
	!
	when error in
		!
		sys_status = sys$asctim(,Sys_buf_22,,)				! get ASCII time into sys_buf_22
!~~~		if (sys_status and 7%) <> 1% then cause error 11		x  not required - call will never fail
		!
		!	transfer data from one map to the other
		!	
		Wcsm_year	= Sys_year					!
!~~~	rset	Wcsm_month	= str$( pos(k_month_names$,Sys_Month,1%) / 3%)	x					bf_100f
		Wcsm_day	= Sys_day					!
		Wcsm_hour	= Sys_hour					!
		Wcsm_minute	= Sys_minute					!
		Wcsm_second	= Sys_second					!
		!
		declare long temp%						!					bf_100f
		temp% = pos(k_month_names$,Sys_Month,1%) / 3%			! compute month number			bf_100f
		if temp% < 10% then						! if less than 10...			bf_100f
		    Wcsm_month_ones	= str$(temp%)				! ...then this goes into ONES		bf_100f
		    Wcsm_month_tens	= "0"					! ...and this goes into TENS		bf_100f
		else								! else >= 10				bf_100f
		    Wcsm_month		= str$(temp%)				!					bf_100f
		end if
		!
		!	make sure there are no spaces in the TENS area of our mapped variables (pad with '0' if necessary)
		!
!~~~		Wcsm_month_tens = "0"	if Wcsm_month_tens	= my_space	x disabled - see above code		bf_100f
		Wcsm_day_tens	= "0"	if Wcsm_day_tens	= my_space	!
		!
		!	now pass result back to caller
		!
		Wcsm_DT_Stamp = Wcsm_Buf_14					! this is it folks
	use
		Wcsm_DT_Stamp = ""						! error so return blank
	end when
	!
	END Function
	!========================================================================================================================
	! Title  : funct_fms_nav
	! Purpose: fms form navigation via terminator
	! Author : NSR
	! Created: 93.05.06
	!========================================================================================================================
32030	function long funct_fms_nav( long	term9%			,&
						direction9%		,&
						fld9%			,&
						field_limit_hi9%	,&
						field_limit_lo9%	,&
						fms_fld_up9()		,&
						fms_fld_dwn9()		)
	    !
	    option type=explicit							!
	    declare long exit_action%							!
	    !
	    !	take some kind of action based upon the entered terminator
	    !
	    exit_action%=1%								! assume loop-to-goto
	    !
	    select term9%
		case 0%									! <ret>
		    exit_action%=0%							!	don't loop
		case 1%,								! <tab>			&
		     3%,								! <auto tab>		&
		     11%,								! <illegal next>	&
		     13%,								! <illegal auto-tab>	&
		     38%,								! next screen		&
		     1030%								! <ctrl-f>
			fld9% = fld9% + 1%						!	advance
			direction9% = 1%						!	forward
	    	case 2%,								! <bs>			&
		     12%,								! <illegal previous>	&
		     35%,								! remove  (p.c. page up)&
		     37%,								! prev screen (p.c. end)&
		     1026%								! <ctrl-b>
			fld9% = fld9% - 1%						!	retreat
			direction9% = -1%						!	backward
		case 17%								! up arrow (illegal up)
			direction9% = -1%						!	backward
			if fms_fld_up9( fld9%) < field_limit_lo9% then
				fld9% = fld9% - 1%					! try retreating
			else
				fld9% = fms_fld_up9(fld9%)				! try navigating
			end if
		case 16%								! down arrow (illegal down)
			direction9% = 1%						!	forward
			if fms_fld_dwn9(fld9%) > field_limit_hi9% then
				fld9% = fld9% + 1%					!	try advancing
			else
				fld9% = fms_fld_dwn9(fld9%)				!	try navigating
			end if
!~
!~	~ for debugging nav logic -
!~
!~			call fdv$putl(	"Fld: "+ str$( fld9% )			+" Dwn: "+ str$( fms_fld_dwn9 (fld9%))	+&
!~					" Up: "+ str$( fms_fld_up9  (fld9%))	+" LoL: "+ str$( field_limit_lo9%)	+&
!~					" HiL: "+ str$( field_limit_hi9% ))
!~			sleep 1
!~
!~~~		case 37%							x prev screen (treat as top)
!~~~			fld9%		= field_limit_lo9%			x	goto 1st field
!~~~			direction9%	= 1%					x	forward (incase LO=disabled)
!~~~		case 38%							x next screen (treat as bottom)
!~~~			fld9%		= field_limit_hi9%			x	goto last field
!~~~			direction9%	= -1%					x	backward (incase HI=disabled)
		case else
			call fdv$putl( "FMS Terminator '"+ str$( term9%) +"' is not currently supported.")
		end select
		!
		funct_fms_nav = exit_action%
		!
	end function
	!
	!	find field name in array
	!
32040	function long find_field_number(string yada() by ref, long array_siz by ref, string target$)
	option type=explicit
	declare long fld
	when error in
	    for fld = 1 to array_siz
		if yada(fld) = target$
		then
		    goto found
		end if
	    next fld
	    found:
	use
	    fld = 0
	end when
	find_field_number = fld
	end function
	!======================================================================= the last line
